/*
     
@memo      
@name Compiler_Check
@author .. , .. 
@version 1.0
@see Compiler
*/


#include "errcode.h"		/*   */
#include "compiler.h"		/*  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


static void PrintQueryTree(treenode_t *); /*    (  ",  ,  ") */

/*
@memo  ,    
*/
#define MYNODE	0

/*
@memo   
*/
#define TEST_QUERY_QTY	4

/*
@memo       
*/
#define QUERY_LINE_MAX	(1<<TREE_LEVEL_MAX)

/*
@memo  
*/
static const char * TEST_QUERY[TEST_QUERY_QTY][QUERY_LINE_MAX]={

	/* JOIN(1, SCAN(0), SCAN(1)) */	
	{ 
	"0 J 1 #0 #1", 
	"",	"", "", "",	"", "", "",	"", "", "",	"", "", "",	"", ""
	},

	/* JOIN(1, JOIN(SCAN(0), SCAN(1)), SCAN(2)) */	
	{ 
	"0 J 1 1 #2", 
	"1 J 1 #0 #1",	
	"", "", "",	"", "", "",	"", "", "",	"", "", "",	"", ""
	},

	/* JOIN(3, JOIN(1, SCAN(2), RESTRICT(1, <, 31, SCAN(2))), RESTRICT(2, >, 19, SCAN(1))) */
	{
	"0 J 3 1 2",
	"1 J 1 #2 3",
	"2 R 2 > 19 #1",
	"3 R 1 < 31 #2",
	"",	"", "", "",	"", "", "",	"", "", "",	"", ""
	},

	/* JOIN(3, JOIN(1, RESTRICT(1, <, 31, SCAN(0)), RESTRICT(2, =, 10, SCAN(1))), JOIN(2, RESTRICT(3, <, 90, SCAN(2)), RESTRICT(0, >, 4, SCAN(3)))) */
	{
	"0 J 3 1 2",
	"1 J 1 3 4",
	"2 J 2 5 6",
	"3 R 1 < 31 #0",
	"4 R 2 = 10 #1",	
	"5 R 3 < 90 #2", 
	"6 R 0 > 4 #2", 
	"",	"", "", "",	"", "", "",	"", ""
	}
};

/*
@memo    ( ",  ,  ")
*/
static const char * TEST_TREE[TEST_QUERY_QTY]={
	/* JOIN(1, SCAN(0), SCAN(1)) */	
	"JOIN 1 [7] SCAN 0 [4] SCAN 1 [4] ",

	/* JOIN(1, JOIN(SCAN(0), SCAN(1)), SCAN(2)) */
	"JOIN 1 [10] JOIN 1 [7] SCAN 0 [4] SCAN 1 [4] SCAN 2 [4] ",

	/* JOIN(3, JOIN(1, SCAN(2), RESTRICT(1, <, 31, SCAN(3))), RESTRICT(2, >, 19, SCAN(1))) */
	"JOIN 3 [10] JOIN 1 [7] SCAN 2 [4] RESTRICT 1 < 31 [4] SCAN 2 [4] RESTRICT 2 > 19 [4] SCAN 1 [4] ",

	/* JOIN(3, JOIN(1, RESTRICT(1, <, 31, SCAN(0)), RESTRICT(2, =, 10, SCAN(1))), JOIN(2, RESTRICT(3, <, 90, SCAN(2)), RESTRICT(0, >, 4, SCAN(3)))) */
	"JOIN 3 [13] JOIN 1 [7] RESTRICT 1 < 31 [4] SCAN 0 [4] RESTRICT 2 = 10 [4] SCAN 1 [4] JOIN 2 [7] RESTRICT 3 < 90 [4] SCAN 2 [4] RESTRICT 0 > 4 [4] SCAN 2 [4] "};

/*
@memo      
*/
static char treestr[1024];



int main(void)
{
	int i, j, res;
	FILE * f;
	treenode_t * tree;

	printf("Test Compiler: START\n");
	printf("\tInit Database and dictionary: START\n");
	res=db_InitDatabase(MYNODE, NULL);
	assert(res==SUCCESS);
	printf("\tInit Database and dictionary: STOP\n");

	printf("\tInit Compiler: START\n");
	res=cp_InitCompiler(MYNODE);
	assert(res==SUCCESS);
	res=cp_InitCompiler(MYNODE); /*   */
	assert(res==SUCCESS);
	printf("\tInit Compiler: STOP\n");
	printf("\tCheck functions to make query tree: START\n");
	for (i=0; i<TEST_QUERY_QTY; i++) {
		printf("\t\tcase #%d of %d: START\n", i+1, TEST_QUERY_QTY);
		f=fopen(QUERY_FNAME, "wt");
		assert(f!=NULL);
		for (j=0; j<QUERY_LINE_MAX; j++)
			if (strlen(TEST_QUERY[i][j])>0)
				fprintf(f, "%s\n", TEST_QUERY[i][j]);
			else
				break;
		fclose(f);
		strcpy(treestr, "");
		tree=cp_CompileQuery(QUERY_FNAME);
		assert(tree!=NULL);
		PrintQueryTree(tree);
		/*
		printf("\nShould be\n'%s'\nIn fact\n'%s'\n", TEST_TREE[i], treestr);
		*/
		assert(strcmp(treestr, TEST_TREE[i])==0);
		printf("\t\tcase #%d of %d: STOP\n", i+1, TEST_QUERY_QTY);
	}
	printf("\tCheck functions to make query tree: STOP\n");


	printf("Test Compiler: STOP\n");

	return 0;
}

/*
    ",  ,  "
@memo   
@param root    
@return   -     ",  ,  " 
(    )
@see treenode_t
*/
static void PrintQueryTree(treenode_t * root)
{
	char buf[10];

	if (root==NULL)
		return; 
	switch (root->operation) {
	case JOIN: 
		strcat(treestr, "JOIN");
		strcat(treestr, " ");
		sprintf(buf, "%d", opJOIN(root, attribute));
		strcat(treestr, buf); 
		strcat(treestr, " ");
		break;
	case RESTRICT: 
		strcat(treestr, "RESTRICT");
		strcat(treestr, " ");
		sprintf(buf, "%d", opRESTRICT(root, attribute));
		strcat(treestr, buf);
		strcat(treestr, " ");
		switch (opRESTRICT(root, theta)) {
		case EQUAL: 
			strcat(treestr, "=");
			break;
		case NOT_EQUAL: 
			strcat(treestr, "<>");
			break;
		case GREATER: 
			strcat(treestr, ">");
			break;
		case LESS: 
			strcat(treestr, "<");
			break;
		case GREATER_OR_EQUAL: 
			strcat(treestr, ">=");
			break;
		case LESS_OR_EQUAL: 
			strcat(treestr, "<=");
			break;
		default: 
			assert(0);
		}
		strcat(treestr, " ");
		sprintf(buf, "%d", opRESTRICT(root, literal));
		strcat(treestr, buf);
		strcat(treestr, " ");
		break;
	case SCAN: 
		strcat(treestr, "SCAN");
		strcat(treestr, " ");
		sprintf(buf, "%d", opSCAN(root, relation));
		strcat(treestr, buf);
		strcat(treestr, " ");
		break;
	default:
		assert(0);
	}
	strcat(treestr, "[");
	sprintf(buf, "%d", root->ary);
	strcat(treestr, buf);
	strcat(treestr, "]");
	strcat(treestr, " ");

	PrintQueryTree(root->left);
	PrintQueryTree(root->right);
}