/*
     .
      :
RELATIONS_IN_DATABASE==4
FRAGMENTATION_ATTRIBUTES=={1, 1, 2, 3}
@memo    
@name AgentManager_Check
@author .. , .. 
@version 1.0
@see AgentManager
*/


#include "errcode.h"	/*   */
#include "compiler.h"	/*   */
#include "lineplan.h"	/*    */
#include "parallel.h"	/*   */
#include "agentmgr.h"	/*    */
#include "filemgr.h"	/*   */
#include "engine.h"		/*  */
#include "dbconfig.h"	/*     */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


/*
@memo   
*/
#define MYNODE	JUNCTION_NODE

/*
@memo       
*/
#define QUERY_LINE_MAX	(1<<TREE_LEVEL_MAX)

/*
@memo  
*/
static const char * TEST_QUERY[QUERY_LINE_MAX]={
	/* SELECT(1, <, 100, SCAN(0)) */
/*
	"0 S 2 > 50 #0",
	"",	"", "", "",	"", "", "",	"", "", "",	"", "", "",	"", ""
*/
	/* JOIN(0, SCAN(0), SCAN(1)) */
/*
	"0 J 0 #0 #1",
	"",	"", "", "",	"", "", "",	"", "", "",	"", "", "",	"", ""
*/
	/* SELECT(0, =, 0, JOIN(0, SCAN(0), SCAN(1)) */
/*	
	"0 S 0 = 0 1",
	"1 J 0 #0 #1",	
	"", "", "",	"", "", "",	"", "", "",	"", "", "",	"", ""
*/
	/* JOIN(0, SELECT(1, >, 10, SCAN(0)), SELECT(1, <, 50, SCAN(1))) */
	"0 J 1 1 2",
	"1 R 1 = 42 #0",	
	"2 R 1 = 42 #1", 
	"", "",	"", "", "",	"", "", "",	"", "", "",	"", ""
};


int main(void)
{
	int i, j, res;
	FILE * f;
	treenode_t * tree;


	printf("Test Engine: START\n");

	printf("\tInit database and dictionary: START\n");
	res=db_InitDatabase(MYNODE, NULL);
	assert(res==SUCCESS);
	printf("\tInit database and dictionary: STOP\n");

	printf("\tInit Compiler: START\n");
	res=cp_InitCompiler(MYNODE);
	assert(res==SUCCESS);
	printf("\tInit Compiler: STOP\n");

	printf("\tInit Plan Generator: START\n");
	res=lp_InitPlanGenerator(MYNODE);
	assert(res==SUCCESS);
	printf("\tInit Plan Generator: STOP\n");

	printf("\tInit Parallelizer: START\n");
	res=pl_InitParallelizer(MYNODE);
	assert(res==SUCCESS);
	printf("\tInit Parallelizer: STOP\n");

	printf("\tInit Agent Manager: START\n");
	res=am_InitAgentManager(MYNODE);
	assert(res==SUCCESS);
	printf("\tInit Agent Manager: STOP\n");

	printf("\tInit Engine: START\n");
	res=en_InitEngine(MYNODE);
	assert(res==SUCCESS);
	res=en_InitEngine(MYNODE); /*   */
	assert(res==SUCCESS);
	printf("\tInit Engine: STOP\n");


	printf("\tInit File Manager: START\n");
	res=fm_InitFileManager(MYNODE);
	printf("\tInit File Manager: STOP\n");
	printf("\tOpen relations: START\n");
	for (i=0; i<RELATIONS_IN_DATABASE; i++) {
		res=fm_OpenRelation(i);
		assert(res==SUCCESS);
	}
	printf("\tOpen relations: STOP\n");

	printf("\tCheck functions to execute query: START\n");
	f=fopen(QUERY_FNAME, "wt");
	assert(f!=NULL);
	for (j=0; j<QUERY_LINE_MAX; j++)
		if (strlen(TEST_QUERY[j])>0)
			fprintf(f, "%s\n", TEST_QUERY[j]);
		else
			break;
	fclose(f);
	
	tree=cp_CompileQuery(QUERY_FNAME);
	assert(tree!=NULL);
	res=lp_GeneratePlan(tree);
	assert(res==SUCCESS);
	
	res=pl_ParallelizeQuery(tree);
	assert(res==SUCCESS);
	
	res=am_MakeAgent(MYNODE, tree);
	assert(res==SUCCESS);

	for (i=0; i<tree->ary; i++)
		printf("A%d\t", i);
	printf("\n");
	res=en_ExecuteQuery(tree);
	assert(res==SUCCESS);
	printf("%d tuple(s).\n", tree->card);
	
	printf("\tCheck functions to execute query: STOP\n");

	printf("\tClose relations: START\n");
	for (i=0; i<RELATIONS_IN_DATABASE; i++) {
		res=fm_CloseRelation(i);
		assert(res==SUCCESS);
	}
	printf("\tClose relations: STOP\n");

	printf("Test Engine: STOP\n");

	return 0;
}
