/*
     
@memo   
@name FileManager_Check
@author .. 
@version 1.0
@see FileManager DatabaseConfig Database
*/

#include "dbconfig.h"	/*     */
#include "dbgen.h"		/*    */
#include "errcode.h"	/*   */
#include "filemgr.h"	/*   */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/*
@memo  ,    
*/
#define MYNODE	0

/*
@memo    
*/
const char * REPLICA_RELATIONS_FILENAMES[RELATIONS_IN_DATABASE]={
	"_R0Fx.txt",
	"_R1Fx.txt",
	"_R2Fx.txt",
	"_R3Fx.txt"
};

int main(void)
{
	int res, i, j, cnt, cnt1, cnt2;
	dictionary_t * D;
	tuple_t tuple, tuple1, tuple2;
	FILE * f1, * f2;

	printf("Test File Manager for node #%d: START\n", MYNODE);

	printf("\tCreate database: START\n");
	res=db_InitDatabase(MYNODE, NULL);
	assert(res==SUCCESS);
	D=db_DatabaseDictionary();
	assert(D!=NULL);
	printf("\tCreate database: STOP\n");

	printf("\tInitialize File System: START\n");
	res=fm_InitFileManager(MYNODE);
	assert(res==SUCCESS);
	res=fm_InitFileManager(MYNODE); /*   */
	assert(res==SUCCESS);
	printf("\tInitialize File System: STOP\n");

	printf("\tCheck functions to open/close relations: START\n");
	res=fm_OpenRelation(RELATIONS_IN_DATABASE+1);
	assert(res!=SUCCESS);
	res=fm_CloseRelation(1);
	assert(res!=SUCCESS);
	res=fm_ResetRelation(RELATIONS_IN_DATABASE-1);
	assert(res!=SUCCESS);
	printf("\tCheck functions to open/close relations: STOP\n");

	printf("\tCheck functions to scan relations: START\n");
	tuple=(tuple_t) calloc(ATTRIBUTES_IN_RELATION, sizeof(domain_t));
	assert(tuple!=NULL);
	tuple1=(tuple_t) calloc(ATTRIBUTES_IN_RELATION, sizeof(domain_t));
	assert(tuple1!=NULL);
	tuple2=(tuple_t) calloc(ATTRIBUTES_IN_RELATION, sizeof(domain_t));
	assert(tuple2!=NULL);
	for (i=0; i<RELATIONS_IN_DATABASE; i++) {
		printf("\t\tScan relation #%d (file %s): START\n", i, D->relations[i].filename);
		res=fm_OpenRelation(i);
		assert(res==SUCCESS);
		res=fm_ResetRelation(i);
		assert(res==SUCCESS);
		f2=fopen(REPLICA_RELATIONS_FILENAMES[i], "wt");
		assert(f2!=NULL);
		cnt=0;
		while (!IsEOF(tuple=fm_NextTuple(i))) {
			assert(tuple!=NULL);
			cnt++;
			for (j=0; j<ATTRIBUTES_IN_RELATION; j++)
				fprintf(f2, "%d\t", tuple[j]);
			fprintf(f2, "\n");
		}
		res=fm_CloseRelation(i);
		assert(res==SUCCESS);
		fclose(f2);

		f1=fopen(RELATIONS_FILENAMES[i][MYNODE], "rt");
		assert(f1!=NULL);
		f2=fopen(REPLICA_RELATIONS_FILENAMES[i], "rt");
		assert(f2!=NULL);
		cnt1=cnt2=0;
		while (!feof(f1) && !feof(f2)) {
			for (j=0; j<ATTRIBUTES_IN_RELATION; j++) {
				fscanf(f1, "%d\t", &tuple1[j]);
				fscanf(f2, "%d\t", &tuple2[j]);
				assert(tuple1[j]==tuple2[j]);
			}
			fscanf(f1, "\n");
			cnt1++;
			fscanf(f2, "\n");
			cnt2++;
		}
		fclose(f1);
		fclose(f2);
		assert(cnt1==cnt2);
		assert(cnt==cnt1);
		printf("\t\tScan relation #%d (file %s): STOP\n", i, D->relations[i].filename);
	}
	printf("\tCheck functions to scan relations: STOP\n");

	printf("Test File Manager for node #%d: STOP\n", MYNODE);

	return 0;
}