/*
      
@memo    
@name MessageManager_Check
@author .. , .. 
@version 1.0
*/

//#undef NDEBUG
#define NDEBUG

#include "msgmgr.h"		/*  */
#include "errcode.h"	/*   */
#include "common.h"		/*     */
#include "debug.h"		/*  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

//#undef NDEBUG
#define NDEBUG


/*
@memo  
*/
#define MSG_LEN	(NODES_IN_CLUSTER)

/*
@memo ,   
*/
#define MASTER_NODE		(0)

/*
@memo  :  
*/
#define MAKE_STMT(stmt)	do { stmt } while (0)

/*
@memo   
*/
#define PRINTF(fmtstr_and_arglist)	\
	MAKE_STMT( if (whoami==MASTER_NODE) { printf fmtstr_and_arglist ; } ; )

/*
@memo   
*/
#ifndef NDEBUG

#define DBG(fmtstr_and_arglist)	\
	MAKE_STMT( printf("<%d> %03d\t", whoami, ++dbgcnt); printf fmtstr_and_arglist ; printf("\n"); )

#else

#define DBG(fmtstr_and_arglist) ((void) 0)

#endif 

/*
@memo   
*/
static int whoami;

static int dbgcnt=0;

/*
     
1)       . 
	 :
		  --  ,    .
		     MSG_LEN.
2)          .
	 :
		  --  ,   -.
		     MSG_LEN
*/
int main(void)
{
	int res, i, j, m, cntSendDone, cntRecvDone;
	int map[NODES_IN_CLUSTER-1];
	int msgSend[NODES_IN_CLUSTER-1][MSG_LEN];
	int msgRecv[NODES_IN_CLUSTER-1][MSG_LEN];
	mid_t midSend[NODES_IN_CLUSTER-1];
	mid_t midRecv[NODES_IN_CLUSTER-1];
	
	res=mm_InitMessageManager();
	assert(res==SUCCESS);
	res=mm_InitMessageManager(); /*   */
	assert(res==SUCCESS);

	res=mm_Synchronize();
	assert(res==SUCCESS);

	whoami=mm_WhoAmI();
	DBG(("START test on %d node", whoami));

	dbg_Init(whoami);

	PRINTF(("Test Message Manager: START\n"));
	m=0;
	for (i=0; i<NODES_IN_CLUSTER; i++) 
		if (i!=whoami) 
			map[m++]=i;
	for (i=0; i<NODES_IN_CLUSTER-1; i++) 
		DBG(("map[%d]==%d", i, map[i]));

	for (i=0; i<NODES_IN_CLUSTER-1; i++) 
		for (j=0; j<MSG_LEN; j++) {
			msgSend[i][j]=whoami;
			msgRecv[i][j]=-whoami;
		}

	DBG(("Run send & receive msg operations from other nodes"));
	for (i=0; i<NODES_IN_CLUSTER-1; i++) {
		DBG(("Running receive message\tfrom node==%d\tvia port==%d\tto buf==%p\tlen==%d", map[i], whoami, &msgRecv[i], sizeof(int)*MSG_LEN));
		midRecv[i]=mm_RunRecv(/*  */ whoami*101+map[i], /*  */ map[i],  &msgRecv[i], sizeof(int)*MSG_LEN);
		DBG(("midRecv==%d", midRecv[i]));
		assert(midRecv>=0);

		DBG(("Running send message\tto node==%d\tvia port==%d\tfrom buf==%p\tlen==%d", map[i], map[i], &msgSend[i], sizeof(int)*MSG_LEN));
		midSend[i]=mm_RunSend(/*  */ whoami+101*map[i], /*  */ map[i], &msgSend[i], sizeof(int)*MSG_LEN);
		DBG(("midSend==%d", midSend[i]));
		assert(midSend>=0);


	}
	DBG(("Done running send & receive msg operations from other nodes"));

	cntSendDone=cntRecvDone=0;

	DBG(("Check either send msg operations is done"));
	while (TRUE) {
		for (i=0; i<NODES_IN_CLUSTER-1; i++) {
			DBG(("Check send msg oper with ID==%d", midSend[i]));
			if (mm_IsDone(midSend[i]))
				cntSendDone++;
			if (cntSendDone>NODES_IN_CLUSTER-1)
				cntSendDone=NODES_IN_CLUSTER-1;
			DBG(("Check recv msg oper with ID==%d", midRecv[i]));
			if (mm_IsDone(midRecv[i]))
				cntRecvDone++;
			if (cntRecvDone>NODES_IN_CLUSTER-1)
				cntRecvDone=NODES_IN_CLUSTER-1;
		}
		if (cntSendDone==NODES_IN_CLUSTER-1 && cntRecvDone==NODES_IN_CLUSTER-1)
			break;
	}
	DBG(("All send & receive msg operations done"));

	res=mm_Synchronize();
	assert(res==SUCCESS);

	for (i=0; i<NODES_IN_CLUSTER-1; i++) {
		DBG(("Received from %d node", map[i]));
		for (j=0; j<MSG_LEN; j++)
			DBG(("msgRecv[%d][%d]==%d", i, j, msgRecv[i][j]));
	}
	
	for (i=0; i<NODES_IN_CLUSTER-1; i++)
		for (j=0; j<MSG_LEN; j++)
			assert(msgRecv[i][j]==map[i]);

	DBG(("Done Msg Manager on %d node", whoami));

	res=mm_DoneMessageManager();
	assert(res==SUCCESS);

	DBG(("STOP test on %d node", whoami));

	PRINTF(("Test Message Manager: STOP\n"));

	return 0;
}