/*
     .
      :
RELATIONS_IN_DATABASE==3
FRAGMENTATION_ATTRIBUTES=={1, 1, 2}
@memo    
@name Parallelizer_Check
@author .. , .. 
@version 1.0
@see Parallelizer Compiler DatabaseConfig RELATIONS_IN_DATABASE FRAGMENTATION_ATTRIBUTES
*/


#include "errcode.h"	/*   */
#include "compiler.h"	/*   */
#include "lineplan.h"	/*    */
#include "parallel.h"	/*  */
#include "dbconfig.h"	/*     */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


/*    (  ",  ,  ") */
static void PrintParallelQueryTree(treenode_t *);

/*
@memo  ,    
*/
#define MYNODE	0

/*
@memo   
*/
#define TEST_QUERY_QTY	4

/*
@memo       
*/
#define QUERY_LINE_MAX	(1<<TREE_LEVEL_MAX)

/*
@memo  
*/
static const char * TEST_QUERY[TEST_QUERY_QTY][QUERY_LINE_MAX]={
	/* JOIN(1, SCAN(0), SCAN(1)) */
	{
	"0 J 1 #0 #1",
	"",	"", "", "",	"", "", "",	"", "", "",	"", "", "",	"", ""
	},

	/* JOIN(1, SCAN(2), SCAN(0)) */
	{
	"0 J 1 #2 #0",
	"",	"", "", "",	"", "", "",	"", "", "",	"", "", "",	"", ""
	},

	/* JOIN(2, JOIN(2, SCAN(1), RESTRICT(1, <, 31, SCAN(2))), RESTRICT(2, >, 19, SCAN(0))) */
	{
	"0 J 2 1 2",
	"1 J 2 #1 3",
	"2 R 2 > 19 #0",
	"3 R 1 < 31 #2",
	"",	"", "", "",	"", "", "",	"", "", "",	"", ""
	},

	/* JOIN(3, JOIN(2, RESTRICT(1, >, 10, SCAN(0)), RESTRICT(2, <, 8, SCAN(2))), JOIN(3, SCAN(2), RESTRICT(1, =, 7, JOIN(3, SCAN(3), RESTRICT(2, < ,4, SCAN(1)))))) */
	{
	"0 J 3 1 2",
	"1 J 2 3 4",
	"2 J 3 #2 5",
	"3 R 1 > 10 #0",
	"4 R 2 < 8 #2",
	"5 R 1 = 7 6",
	"6 J 3 #2 7",
	"7 R 2 < 4 #1",	
	"", "", "",	"", "", "",	"", ""
	}
};

/*
@memo     ( ",  ,  ")
*/
static const char * TEST_PLAN[TEST_QUERY_QTY]={
	/* JOIN(1, SCAN(0), SCAN(1)) */
	"{JOIN 1} {SCAN 0} {SCAN 1} ",

	/* JOIN(1, SCAN(2), SCAN(0)) */
	"{JOIN 1} {MERGE} {GATHER 1} {SPLIT} {SCAN 2} {SCATTER} {SCAN 0} ",
	
	/* JOIN(2, JOIN(2, SCAN(1), RESTRICT(1, <, 31, SCAN(2))), RESTRICT(2, >, 19, SCAN(0))) */
	"{JOIN 2} {JOIN 2} {MERGE} {GATHER 1} {SPLIT} {SCAN 1} {SCATTER} {RESTRICT 1 < 31} {SCAN 2} {MERGE} {GATHER 2} {SPLIT} {RESTRICT 2 > 19} {SCAN 0} {SCATTER} ",

	/* JOIN(3, JOIN(2, RESTRICT(1, >, 10, SCAN(0)), RESTRICT(2, <, 8, SCAN(2))), JOIN(3, SCAN(2), RESTRICT(1, =, 7, JOIN(3, SCAN(3), RESTRICT(2, < ,4, SCAN(1)))))) */
	"{JOIN 3} {MERGE} {GATHER 2} {SPLIT} {JOIN 2} {MERGE} {GATHER 1} {SPLIT} {RESTRICT 1 > 10} {SCAN 0} {SCATTER} {RESTRICT 2 < 8} {SCAN 2} {SCATTER} {JOIN 3} {MERGE} {GATHER 3} {SPLIT} {SCAN 2} {SCATTER} {RESTRICT 1 = 7} {JOIN 3} {MERGE} {GATHER 4} {SPLIT} {SCAN 2} {SCATTER} {MERGE} {GATHER 5} {SPLIT} {RESTRICT 2 < 4} {SCAN 1} {SCATTER} "
};

/*
@memo      
*/
static char treestr[1024];



int main(void)
{
	int i, j, res;
	FILE * f;
	treenode_t * tree;

	printf("Test Parallelizer: START\n");

	printf("\tInit database and dictionary: START\n");
	res=db_InitDatabase(MYNODE, NULL);
	assert(res==SUCCESS);
	printf("\tInit database and dictionary: STOP\n");

	printf("\tInit Compiler: START\n");
	res=cp_InitCompiler(MYNODE);
	assert(res==SUCCESS);
	printf("\tInit Compiler: STOP\n");

	printf("\tInit Plan Generator: START\n");
	res=lp_InitPlanGenerator(MYNODE);
	assert(res==SUCCESS);
	printf("\tInit Plan Generator: STOP\n");

	printf("\tInit Parallelizer: START\n");
	res=pl_InitParallelizer(MYNODE);
	assert(res==SUCCESS);
	res=pl_InitParallelizer(MYNODE); /*   */
	assert(res==SUCCESS);
	printf("\tInit Parallelizer: STOP\n");

	printf("\tCheck functions to make parallel query tree: START\n");
	for (i=0; i<TEST_QUERY_QTY; i++) {
		printf("\t\tcase #%d of %d: START\n", i+1, TEST_QUERY_QTY);
		f=fopen(QUERY_FNAME, "wt");
		assert(f!=NULL);
		for (j=0; j<QUERY_LINE_MAX; j++)
			if (strlen(TEST_QUERY[i][j])>0)
				fprintf(f, "%s\n", TEST_QUERY[i][j]);
			else
				break;
		fclose(f);
		strcpy(treestr, "");
		tree=cp_CompileQuery(QUERY_FNAME);
		assert(tree!=NULL);

		res=lp_GeneratePlan(tree);
		assert(res==SUCCESS);

		res=pl_ParallelizeQuery(tree);
		assert(res==SUCCESS);

		PrintParallelQueryTree(tree);
/*		
		printf("\nShould be\n'%s'\nIn fact\n'%s'\n", TEST_PLAN[i], treestr);
*/		
		assert(strcmp(treestr, TEST_PLAN[i])==0);
		printf("\t\tcase #%d of %d: STOP\n", i+1, TEST_QUERY_QTY);
	}
	printf("\tCheck functions to make parallel query tree: STOP\n");

	printf("Test Parallelizer: STOP\n");

	return 0;
}


/*
@memo     
@param root    
*/
static void PrintParallelQueryTree(treenode_t * root)
{
	char buf[10];

	if (root==NULL)
		return;
	strcat(treestr, "{");
	switch (root->operation) {
	case JOIN:
		strcat(treestr, "JOIN");
		strcat(treestr, " ");
		sprintf(buf, "%d", opJOIN(root, attribute));
		strcat(treestr, buf);
		break;
	case RESTRICT:
		strcat(treestr, "RESTRICT");
		strcat(treestr, " ");
		sprintf(buf, "%d", opRESTRICT(root, attribute));
		strcat(treestr, buf);
		strcat(treestr, " ");
		switch (opRESTRICT(root, theta)) {
		case EQUAL: 
			strcat(treestr, "=");
			break;
		case NOT_EQUAL: 
			strcat(treestr, "<>");
			break;
		case GREATER: 
			strcat(treestr, ">");
			break;
		case LESS: 
			strcat(treestr, "<");
			break;
		case GREATER_OR_EQUAL: 
			strcat(treestr, ">=");
			break;
		case LESS_OR_EQUAL: 
			strcat(treestr, "<=");
			break;
		default: 
			assert(0);
		}
		strcat(treestr, " ");
		sprintf(buf, "%d", opRESTRICT(root, literal));
		strcat(treestr, buf);
		break;
	case SCAN: 
		strcat(treestr, "SCAN");
		strcat(treestr, " ");
		sprintf(buf, "%d", opSCAN(root, relation));
		strcat(treestr, buf);
		break;
	case MERGE:
		strcat(treestr, "MERGE");
		assert(root->executefn==ExecuteMERGE);
		assert(root->resetfn==ResetMERGE);
		assert(root->left->operation==GATHER);
		assert(root->right->operation==SPLIT);
		break;
	case GATHER:
		strcat(treestr, "GATHER");
		strcat(treestr, " ");
		sprintf(buf, "%d", opGATHER(root, port));
		strcat(treestr, buf);
		assert(root->executefn==ExecuteGATHER);
		assert(root->resetfn==ResetGATHER);
		assert(root->left==NULL);
		assert(root->right==NULL);
		break;
	case SPLIT:
		strcat(treestr, "SPLIT");
		assert(root->executefn==ExecuteSPLIT);
		assert(root->resetfn==ResetSPLIT);
		assert(root->right->operation==SCATTER);
		assert(root->left->operation==JOIN || root->left->operation==RESTRICT || root->left->operation==SCAN);
		break;
	case SCATTER:
		strcat(treestr, "SCATTER");
		assert(root->executefn==ExecuteSCATTER);
		assert(root->resetfn==ResetSCATTER);		
		assert(root->left==NULL);
		assert(root->right==NULL);
		break;
	default:
		assert(0);
	}

	strcat(treestr, "} ");
	PrintParallelQueryTree(root->left);
	PrintParallelQueryTree(root->right);
}


