#ifndef _PARALLEL_H
#define _PARALLEL_H

/**
      
          
.       EXCHANGE.
@memo  
@name  Parallelizer
@author (c) 2004 .. , .. 
@version 1.0  \today
@see PlanGenerator AgentManager MessageManager 
*/

#include "lineplan.h"	/*    */
#include "msgmgr.h"		/*   */

/*@{*/

/**
    
@name Parallelizer_Types
@memo   
*/

/*@{*/

typedef int (* _distribute_fn_t)(tuple_t, int);

/**
   EXCHANGE   ,    
  .

  :
\begin{verbatim}
typedef int (* _distribute_fn_t)(tuple_t, int);
\end{verbatim}
    :

\begin{tabular}{lll}
{\bf N / }	& {\bf 	} & {\bf  } \\
1		& \Ref{tuple_t}	&    \\
2		& int		&  ,     --    0  \Ref{ATTRIBUTES_IN_RELATION}-1 
\end{tabular}

     int --   (..  
 ,       )    0  
\Ref{FRAGMENTS_IN_RELATION}-1 .
@memo      EXCHANGE
*/
typedef _distribute_fn_t distribute_fn_t;

struct _opEXCHANGE {
	int port;
	distribute_fn_t distributefn;
};

/*
  EXCHANGE ()   :
\begin{verbatim}
int port 						-   
distribute_fn_t	distributefn	-    
\end{verbatim}
@memo   EXCHANGE
@see distribute_fn_t opJOIN_t opRESTRICT_t opSCAN_t opSTORE_t
*/
typedef struct _opEXCHANGE opEXCHANGE_t;


struct _opMERGE {
	int cnt;
};

/**
    MERGE (   EXCHANGE)
.
@memo   MERGE
@see opGATHER_t opSCATTER_t opSPLIT_t
*/
typedef struct _opMERGE opMERGE_t;

struct _opSPLIT {
	distribute_fn_t distributefn;
	int breakover;
	int cntEOF;
};

/**
    SPLIT (   EXCHANGE) 
.
@memo   SPLIT
@see opGATHER_t opSCATTER_t opMERGE_t
*/
typedef struct _opSPLIT opSPLIT_t;

struct _opSCATTER {
	int port;
	mid_t midSend[NODES_IN_CLUSTER];
	tuple_t bufSend[NODES_IN_CLUSTER];
	int ready[NODES_IN_CLUSTER];
};

/**
    SCATTER (   EXCHANGE) 
.
@memo   SCATTER
@see opGATHER_t opSPLIT_t opMERGE_t
*/
typedef struct _opSCATTER opSCATTER_t;

struct _opGATHER {
	int port;
	mid_t midRecv[NODES_IN_CLUSTER];
	tuple_t bufRecv[NODES_IN_CLUSTER];
	int cntEOF;
};

/**
    GATHER (   EXCHANGE) 
.
@memo   SCATTER
@see opSPLIT_t opSCATTER_t opMERGE_t NODES_IN_CLUSTER
*/
typedef struct _opGATHER opGATHER_t;

/*@}*/

/**
    
@name Parallelizer_Functions
@memo   
*/

/*@{*/

/**
       
         .
       .
@memo   
@param node	   	
@return 0       
@see pl_ParallelizeQuery ErrorCodes
*/
extern int pl_InitParallelizer(int node);

/**
        
    .     EXCHANGE 
" "    .   0   
 EXCHANGE,     .
@memo  
@param root        
@return 0       
@see treenode_t AgentManager am_MakeAgent ErrorCodes
*/
extern int pl_ParallelizeQuery(treenode_t * root);

/**
     MERGE (  EXCHANGE). 
        " ".
@memo   MERGE
@param node     
@return       MERGE;     
 -EOF
@see Bracket EOFtuple execute_fn_t ResetMERGE ExecuteSPLIT ExecuteSCATTER ExecuteGATHER ExecuteSTORE ExecuteRESTRICT ExecuteJOIN_NestedLoops
*/
extern tuple_t ExecuteMERGE(treenode_t * node);

/**
   MERGE      .
@memo   MERGE
@param node     
@return 0       
@see reset_fn_t ExecuteMERGE ErrorCodes
*/
extern int ResetMERGE(treenode_t * node);

/**
     SPLIT (  EXCHANGE). 
      " ".
@memo   SPLIT
@param node     
@return       SPLIT     
NULL      
@see Bracket execute_fn_t ResetSPLIT ExecuteMERGE ExecuteSCATTER ExecuteGATHER ExecuteSTORE ExecuteRESTRICT ExecuteJOIN_NestedLoops
*/
extern tuple_t ExecuteSPLIT(treenode_t * node);

/**
   SPLIT      .
@memo   SPLIT
@param node     
@return 0       
@see reset_fn_t ExecuteSPLIT ErrorCodes
*/
extern int ResetSPLIT(treenode_t * node);

/**
     GATHER (  EXCHANGE). 
      " "   
 .
@memo   GATHER
@param node     
@return       GATHER    (-EOF
  )  NULL      
@see Bracket MessageManager execute_fn_t EOFtuple ResetGATHER ExecuteSPLIT ExecuteSCATTER ExecuteMERGE ExecuteSTORE ExecuteRESTRICT ExecuteJOIN_NestedLoops
*/
extern tuple_t ExecuteGATHER(treenode_t * node);

/**
  GATHER          
   ()     .  
       .
@memo   GATHER
@param node     
@return 0       
@see MessageManager reset_fn_t ExecuteGATHER ErrorCodes
*/
extern int ResetGATHER(treenode_t * node);

/**
     SCATTER (  EXCHANGE). 
      " "   
 .
@memo   SCATTER
@param node     
@return  NULL
@see Bracket MessageManager execute_fn_t ExecuteSPLIT ExecuteMERGE ExecuteGATHER ExecuteSTORE ExecuteRESTRICT ExecuteJOIN_NestedLoops
*/
extern tuple_t ExecuteSCATTER(treenode_t * node);

/**
   SCATTER     ,     
  : return SUCCESS;
@memo   SCATTER
@param node     
@return 0       
@see reset_fn_t ExecuteSCATTER ErrorCodes
*/
extern int ResetSCATTER(treenode_t * node);

/*
@memo     EXCHANGE
*/
#define opEXCHANGE(treenode, attr)	((opEXCHANGE_t *) treenode->cargo)->attr

/*
@memo     MERGE
*/
#define opMERGE(treenode, attr)	((opMERGE_t *) treenode->cargo)->attr

/*
@memo     GATHER
*/
#define opGATHER(treenode, attr)	((opGATHER_t *) treenode->cargo)->attr

/*
@memo     SPLIT
*/
#define opSPLIT(treenode, attr)	((opSPLIT_t *) treenode->cargo)->attr

/*
@memo     SCATTER
*/
#define opSCATTER(treenode, attr)	((opSCATTER_t *) treenode->cargo)->attr

/*@}*/

/*@}*/


#endif /* _PARALLEL_H */