/*
       
           
   .
        ,   
   EXCHANGE       -
""   STORE    .  STORE 
   -"".
      STORE.
@memo   
@name  AgentManager_Implementation
@author (c) 2004 .. , .. 
@version 1.0
@see Parallelizer Bracket
*/


#include "agentmgr.h"	/*  */
#include "errcode.h"	/*   */
#include "dbgen.h"		/*    */
#include "bracket.h"	/*   */
#include "msgmgr.h"		/*   */
#include "debug.h"		/*    */
#include "common.h"		/*     */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>


static void PrintAgent(treenode_t *);

/*
@memo     
*/
char agentstr[1024];



/*
@memo     
@param node    
@param tuple   
*/
#define RESCPY(node, tuple)	memcpy(node->result, tuple, node->ary*sizeof(domain_t))

/*@{*/

/*
      
@name AgentManager_Implementation_Variables
@memo     
@see AgentManager_Implementation_Functions
*/

/*@{*/

/*
@memo  
*/
static dictionary_t * DICTIONARY;

/*
@memo   
*/
static int initflag=FALSE;

/*
@memo  
*/
static treenode_t * PARALLEL_AGENT;

/*@}*/


/*
      
@name AgentManager_Implementation_Functions
@memo     
@see AgentManager_Implementation_Variables
*/

/*@{*/

/*
        
         .
       .
@memo    
@param node	   	
@return SUCCESS       
@see am_ParallelizeQuery
*/
int am_InitAgentManager(int node)
{
	DBG_INIT(node);
	START("am_InitAgentManager");
	if (initflag)
		return SUCCESS;
	if (node<0 || node>=FRAGMENTS_IN_RELATION)
		return EINVAL;
	if ((DICTIONARY=db_DatabaseDictionary())==NULL)
		return EINVAL;
	if (node!=DICTIONARY->whoami)
		return EINVAL;
	initflag=1;
	PARALLEL_AGENT=NULL;
	STOP("am_InitAgentManager");
	return SUCCESS;
}

/**
          
     .
@memo   
@param root        
@return        NULL   
@see treenode_t
*/
treenode_t * am_MakeAgent(treenode_t * root)
{
	treenode_t *Store, *Merge, *Gather, *Split, *Scatter;

	START("am_MakeAgent");
	if (!initflag)
		return NULL;
	if (root==NULL)
		return NULL;

	/*  MERGE */
	Merge=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Merge!=NULL);
	Merge->operation=MERGE;
	Merge->cargo=(opMERGE_t *) calloc(1, sizeof(opMERGE_t));
	assert(Merge->cargo!=NULL);
	Merge->executefn=(execute_fn_t) ExecuteMERGE;
	Merge->resetfn=(reset_fn_t) ResetMERGE;
	Merge->ary=root->ary;
	Merge->fragmentattr=root->fragmentattr;
	Merge->fragmentfn=root->fragmentfn;

	/*  GATHER */
	Gather=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Gather!=NULL);
	Gather->operation=GATHER;
	Gather->cargo=(opGATHER_t *) calloc(1, sizeof(opGATHER_t));
	assert(Gather->cargo!=NULL);
	opGATHER(Gather, port)=0; /* !!   >=1    */
	Gather->executefn=(execute_fn_t) ExecuteGATHER;
	Gather->resetfn=(reset_fn_t) ResetGATHER;
	Gather->ary=root->ary;
	Gather->fragmentattr=root->fragmentattr;
	Gather->fragmentfn=root->fragmentfn;

	/*  SPLIT */
	Split=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Split!=NULL);
	Split->operation=SPLIT;
	Split->cargo=(opSPLIT_t *) calloc(1, sizeof(opSPLIT_t));
	assert(Split->cargo!=NULL);
	opSPLIT(Split, distributefn)=am_PMergeFn;
	Split->executefn=(execute_fn_t) ExecuteSPLIT;
	Split->resetfn=(reset_fn_t) ResetSPLIT;
	Split->ary=root->ary;
	Split->fragmentattr=root->fragmentattr;
	Split->fragmentfn=root->fragmentfn;

	/*  SCATTER */
	Scatter=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Scatter!=NULL);
	Scatter->operation=SCATTER;
	Scatter->cargo=(opSCATTER_t *) calloc(1, sizeof(opSCATTER_t));
	assert(Scatter->cargo!=NULL);
	opSCATTER(Scatter, port)=0; /* !!   >=1    */
	Scatter->executefn=(execute_fn_t) ExecuteSCATTER;
	Scatter->resetfn=(reset_fn_t) ResetSCATTER;
	Scatter->ary=root->ary;
	Scatter->fragmentattr=root->fragmentattr;
	Scatter->fragmentfn=root->fragmentfn;
		
	/*  EXCHANGE */
	Merge->left=Gather;
	Merge->right=Split;
	Gather->left=Gather->right=NULL;
	Split->left=root;
	Split->right=Scatter;
	Scatter->left=Scatter->right=NULL;

	/*   */
	Store=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Store!=NULL);
	Store->operation=STORE;
	Store->executefn=(execute_fn_t) ExecuteSTORE;
	Store->resetfn=(reset_fn_t) ResetSTORE;
	Store->ary=root->ary;
	Store->fragmentattr=root->fragmentattr;
	Store->fragmentfn=root->fragmentfn;
	Store->right=NULL;
	Store->left=Merge;
	PARALLEL_AGENT=Store;

	PrintAgent(PARALLEL_AGENT);

	STOP("am_MakeAgent");
	return PARALLEL_AGENT;
}

/*
     :
return JUNCTION_NODE;
@memo     
@param tuple   ;
@param attr       (  0);
		      0     
          .
@return  -""   0..FRAGMENTS_IN_RELATION-1
@see JUNCTION_NODE distribute_fn_t tuple_t
*/
int am_PMergeFn(tuple_t tuple, int attr)
{
	return PMERGE_NODE;
}


/*
     STORE.     
  " ".       stdout.
@memo   STORE
@return       STORE  , 
   --     NULL   
@see execute_fn_t ExecuteRESTRICT ExecuteJOIN Bracket 
*/
tuple_t ExecuteSTORE(treenode_t * node)
{
	int i;

	START("ExecuteSTORE");
	while (!IsEOF(node->result=br_Next(node->left))) {
		assert(node->result!=NULL);
#ifndef NDEBUG
		printf("<%d>\tRESULT\t", DICTIONARY->whoami);
#endif 
		for (i=0; i<node->ary; i++)
			printf("%d\t", node->result[i]);
		printf("\n");
	}
	STOP("ExecuteSTORE");
	return node->result;
}

/*
   STORE     ,     
  : return SUCCESS;
@memo   STORE
@param node     
@return SUCCESS       
@see reset_fn_t ExecuteSTORE
*/
int ResetSTORE(treenode_t * node)
{
	START("ResetSTORE");
	node->result=(tuple_t) calloc(node->ary, sizeof(domain_t));
	assert(node->result!=NULL);
	STOP("ResetSTORE");
	return SUCCESS;
}


/*
@memo   
@param root    
*/
static void PrintAgent(treenode_t * root)
{
	char buf[10];

	if (root==NULL)
		return;
	strcat(agentstr, "{ary==");
	sprintf(buf, "%d", root->ary);
	strcat(agentstr, buf);
	strcat(agentstr, " ");
	assert(root->ary>=ATTRIBUTES_IN_RELATION);
	switch (root->operation) {
	case JOIN:
		strcat(agentstr, "JOIN");
		strcat(agentstr, " ");
		sprintf(buf, "%d", opJOIN(root, attribute));
		strcat(agentstr, buf);
		strcat(agentstr, "} ");
		break;
	case RESTRICT:
		strcat(agentstr, "RESTRICT");
		strcat(agentstr, " ");
		sprintf(buf, "%d", opRESTRICT(root, attribute));
		strcat(agentstr, buf);
		strcat(agentstr, " ");
		switch (opRESTRICT(root, theta)) {
		case EQUAL: 
			strcat(agentstr, "=");
			break;
		case NOT_EQUAL: 
			strcat(agentstr, "<>");
			break;
		case GREATER: 
			strcat(agentstr, ">");
			break;
		case LESS: 
			strcat(agentstr, "<");
			break;
		case GREATER_OR_EQUAL: 
			strcat(agentstr, ">=");
			break;
		case LESS_OR_EQUAL: 
			strcat(agentstr, "<=");
			break;
		default: 
			assert(0);
		}
		strcat(agentstr, " ");
		sprintf(buf, "%d", opRESTRICT(root, literal));
		strcat(agentstr, buf);
		strcat(agentstr, "} ");
		break;
	case SCAN: 
		strcat(agentstr, "SCAN");
		strcat(agentstr, " ");
		sprintf(buf, "%d", opSCAN(root, relation));
		strcat(agentstr, buf);
		strcat(agentstr, "} ");
		break;
	case MERGE:
		strcat(agentstr, "MERGE");
		strcat(agentstr, "} ");
		assert(root->executefn==(execute_fn_t)ExecuteMERGE);
		assert(root->left->operation==GATHER);
		assert(root->right->operation==SPLIT);
		break;
	case GATHER:
		strcat(agentstr, "GATHER");
		strcat(agentstr, " ");
		sprintf(buf, "%d", opGATHER(root, port));
		strcat(agentstr, buf);
		strcat(agentstr, "} ");
		assert(root->executefn==(execute_fn_t)ExecuteGATHER);
		assert(root->left==NULL);
		assert(root->right==NULL);
		break;
	case SPLIT:
		strcat(agentstr, "SPLIT");
		strcat(agentstr, "} ");
		assert(root->executefn==(execute_fn_t)ExecuteSPLIT);
		assert(root->right->operation==SCATTER);
		assert(root->left->operation==JOIN || root->left->operation==RESTRICT || root->left->operation==SCAN);
		break;
	case SCATTER:
		strcat(agentstr, "SCATTER");
		strcat(agentstr, "} ");
		assert(root->executefn==(execute_fn_t)ExecuteSCATTER);
		assert(root->left==NULL);
		assert(root->right==NULL);
		break;
	case STORE:
		strcat(agentstr, "STORE");
		strcat(agentstr, "} ");
		break;
	default:
		assert(0);
	}
	PrintAgent(root->left);
	PrintAgent(root->right);
}

/*@}*/

/*@}*/


