/*
       
        .
     .

.
            .
     ( ):
_		::= __ ENTER [{_  ENTER}]
__::= __ _
__	::= __
_	::= _JOIN | _RESTRICT
_JOIN		::= J _  
_RESTRICT 	::= S _ _  
_		::= __
_::= > | < | =
				::= __
			::= __ | #_
__	::= __
_	::= __
 :
1. _		    0  ATTRIBUTES_IN_RELATION-1 .
2. _		    0  RELATIONS_IN_DATABASE-1 .
3. __	    0  ___-1 .
     :
	   [0 JOIN(=3)]
	  /					  \
	[1 JOIN(=1)]	  [2 RESTRICT(=2, =">", =19)]
	/				  \											   |
=2	[3 RESTRICT(=1, ="<", =31)]   =1
					|
				=3
  :
0 J 3 1 2
1 J 1 #2 3
2 S 2 > 19 #1
3 S 1 < 31 #3

.
        :
::=_;
_	::= _JOIN | _RESTRICT
_JOIN		::= JOIN(_, , )
_RESTRICT 	::= RESTRICT(_, _, , )
_		::= __
_::= > | < | = | >= | <= | <>
				::= __
			::= _
@memo  
@name  Compiler_Implementation
@author .. , .. 
@version 1.0
@see Database DatabaseConfig Compiler
*/


#include "compiler.h"	/*  */
#include "errcode.h"	/*   */
#include "common.h"		/*     */
#include "debug.h"		/*  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


static void MakeQueryTree(char *);					/*    */
static int CalculateAry(treenode_t *);				/*   "" */
static treenode_t * treedup(treenode_t *);


/*@{*/


/*
    
@name Compiler_Implementation_Types
@memo   
@see Compiler_Functions Compiler_Constants
*/

/*@{*/

/*
@memo    
*/
typedef treenode_t * queryarray_t;

/*@}*/

/*
    
@name Compiler_Implementation_Types
@memo   
@see Compiler_Functions Compiler_Constants
*/

/*@{*/

/*
@memo    
*/
static queryarray_t QUERYARRAY;


/*
@memo  
*/
static treenode_t * QUERYTREE;

/*
@memo  
*/
static dictionary_t * DICTIONARY;

/*
@memo   
*/
static int initflag=FALSE;

/*@}*/


/*
    
@name Compiler_Implementation_Functions
@memo   
@see Compiler_Types Compiler_Constants
*/

/*@{*/

/*
       
         .
       .
@memo   
@param node	   	
@return SUCCESS       
@see cp_CompileQuery
*/
int cp_InitCompiler(int node)
{
	DBG_INIT(node);
	START("cp_InitCompiler");
	if (initflag)
		return SUCCESS;
	if (node<0 || node>=FRAGMENTS_IN_RELATION)
		return EINVAL;
	if ((DICTIONARY=db_DatabaseDictionary())==NULL)
		return EINVAL;
	if (node!=DICTIONARY->whoami)
		return EINVAL;
	initflag=TRUE;   
	STOP("cp_InitCompiler");
	return SUCCESS;
}


/*
          
      .
       ""     .
@memo  
@param query     (    -  )
@return          NULL   
@see QUERY_FNAME
*/
treenode_t * cp_CompileQuery(char * query)
{
	START("cp_CompileQuery");
	if (!initflag)
		return NULL;
	MakeQueryTree(query);
	CalculateAry(QUERYTREE);
	STOP("cp_CompileQuery");
	return QUERYTREE;
}


/*
@memo   
*/
static void MakeQueryTree(char * query)
{
	FILE * f;
	char c;
	int i, nodeqty, noleaves, leafqty, attr, r1, r2, literal;

	START("MakeQueryTree");
	nodeqty=(1<<(TREE_LEVEL_MAX+1))-1;	/*   ,   -   SCAN */
	noleaves=(1<<TREE_LEVEL_MAX)-1;		/*  ,    -   SCAN */

	QUERYARRAY=(treenode_t *) calloc(nodeqty, sizeof(treenode_t));
	assert(QUERYARRAY!=NULL);
	for (i=0; i<nodeqty; i++) {
		QUERYARRAY[i].fragmentattr=NIL;
		QUERYARRAY[i].cargo=NULL;	
		QUERYARRAY[i].fragmentfn=NULL;
		QUERYARRAY[i].ary=NIL;
		QUERYARRAY[i].card=0;
		QUERYARRAY[i].executefn=NULL;
		QUERYARRAY[i].resetfn=NULL;
		QUERYARRAY[i].result=NULL;
		QUERYARRAY[i].isBOF=QUERYARRAY[i].isEOF=NIL;
		QUERYARRAY[i].left=QUERYARRAY[i].right=NULL;
	}

	leafqty=0;
	f=fopen(query, "rt");
	while (!feof(f)) {
		fscanf(f, "%d %c %d ", &i, &c, &attr);
		if (c=='j' || c=='J') {
			QUERYARRAY[i].operation=JOIN;
			QUERYARRAY[i].cargo=(opJOIN_t *) calloc(1, sizeof(opJOIN_t));
			assert(QUERYARRAY[i].cargo!=NULL);
			((opJOIN_t *) QUERYARRAY[i].cargo)->attribute=attr;

			fscanf(f, "%c", &c);
			if (c=='#') {
				fscanf(f, "%d", &r1);
				QUERYARRAY[noleaves+leafqty].operation=SCAN;
				QUERYARRAY[noleaves+leafqty].cargo=(opSCAN_t *) calloc(1, sizeof(opSCAN_t));
				assert(QUERYARRAY[noleaves+leafqty].cargo!=NULL);
				((opSCAN_t *) QUERYARRAY[noleaves+leafqty].cargo)->relation=r1;
				QUERYARRAY[noleaves+leafqty].left=QUERYARRAY[noleaves+leafqty].right=NULL;
				QUERYARRAY[i].left=&QUERYARRAY[noleaves+leafqty];
				leafqty++;
			} else
				QUERYARRAY[i].left=&QUERYARRAY[c-'0'];
			fscanf(f, " %c", &c);
			if (c=='#') {
				fscanf(f, "%d", &r2);
				QUERYARRAY[noleaves+leafqty].operation=SCAN;
				QUERYARRAY[noleaves+leafqty].cargo=(opSCAN_t *) calloc(1, sizeof(opSCAN_t));
				assert(QUERYARRAY[noleaves+leafqty].cargo!=NULL);
				((opSCAN_t *) QUERYARRAY[noleaves+leafqty].cargo)->relation=r2;
				QUERYARRAY[noleaves+leafqty].left=QUERYARRAY[noleaves+leafqty].right=NULL;
				QUERYARRAY[i].right=&QUERYARRAY[noleaves+leafqty];
				leafqty++;
			} else
				QUERYARRAY[i].right=&QUERYARRAY[c-'0'];
		} else
			if (c=='r' || c=='R') {
				QUERYARRAY[i].operation=RESTRICT;
				QUERYARRAY[i].cargo=(opRESTRICT_t *) calloc(1, sizeof(opRESTRICT_t));
				assert(QUERYARRAY[i].cargo!=NULL);
				QUERYARRAY[i].right=NULL;
				((opRESTRICT_t *) QUERYARRAY[i].cargo)->attribute=attr;
				fscanf(f, "%c %d ", &c, &literal);
				switch (c) {
				case '>': ((opRESTRICT_t *) QUERYARRAY[i].cargo)->theta=GREATER; break;
				case '<': ((opRESTRICT_t *) QUERYARRAY[i].cargo)->theta=LESS; break;
				case '=': ((opRESTRICT_t *) QUERYARRAY[i].cargo)->theta=EQUAL; break;
				default : assert(0);
				}
				((opRESTRICT_t *) QUERYARRAY[i].cargo)->literal=literal;
				fscanf(f, " %c", &c);
				if (c=='#') {
					fscanf(f, "%d", &r1);
					QUERYARRAY[noleaves+leafqty].operation=SCAN;
					QUERYARRAY[noleaves+leafqty].cargo=(opSCAN_t *) calloc(1, sizeof(opSCAN_t));
					assert(QUERYARRAY[noleaves+leafqty].cargo!=NULL);
					((opSCAN_t *) QUERYARRAY[noleaves+leafqty].cargo)->relation=r1;
					QUERYARRAY[noleaves+leafqty].left=QUERYARRAY[noleaves+leafqty].right=NULL;
					QUERYARRAY[i].left=&QUERYARRAY[noleaves+leafqty];
					leafqty++;
				} else
					QUERYARRAY[i].left=&QUERYARRAY[c-'0'];
			} else
				assert(0);
		fscanf(f, "\n");
	}

	QUERYTREE=treedup(&QUERYARRAY[0]);
	STOP("MakeQueryTree");
}


/*
@memo   ""
*/
static int CalculateAry(treenode_t * root)
{
	int left_ary, right_ary;

	START("CalculateAry");
	if (root==NULL) {
		RETURN_INT("CalculateAry", 0);
		STOP("CalculateAry");
		return 0;
	}
	switch (root->operation) {
	case SCAN:		
		assert(opSCAN(root, relation)>=0);
		assert(opSCAN(root, relation)<RELATIONS_IN_DATABASE);
		root->ary=DICTIONARY->relations[opSCAN(root, relation)].ary;
		assert(root->ary==ATTRIBUTES_IN_RELATION);
		break;
	case RESTRICT:
		root->ary=CalculateAry(root->left);
		break;
	case JOIN:
		left_ary=CalculateAry(root->left);
		right_ary=CalculateAry(root->right);
		root->ary=left_ary+right_ary-1;
		break;
	default:
		assert(0);
	}
	RETURN_INT("CalculateAry", root->ary);
	STOP("CalculateAry");
	return root->ary;
}

static treenode_t * treedup(treenode_t * t1)
{
	treenode_t * t2;

	if (t1==NULL)
		return NULL;
	t2=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(t2!=NULL);
	t2->ary=t1->ary;
	t2->card=t1->card;
	t2->executefn=t1->executefn;
	t2->fragmentattr=t1->fragmentattr;
	t2->fragmentfn=t1->fragmentfn;
	t2->resetfn=t1->resetfn;
	t2->result=t1->result;
	t2->operation=t1->operation;
	switch (t2->operation) {
	case JOIN:
		t2->cargo=(opJOIN_t *) calloc(1, sizeof(opJOIN_t));
		assert(t2->cargo!=NULL);
		opJOIN(t2, attribute)=opJOIN(t1, attribute);
		break;
	case RESTRICT:
		t2->cargo=(opRESTRICT_t *) calloc(1, sizeof(opRESTRICT_t));
		assert(t2->cargo!=NULL);
		opRESTRICT(t2, attribute)=opRESTRICT(t1, attribute);
		opRESTRICT(t2, theta)=opRESTRICT(t1, theta);
		opRESTRICT(t2, literal)=opRESTRICT(t1, literal);
		break;
	case SCAN:
		t2->cargo=(opSCAN_t *) calloc(1, sizeof(opSCAN_t));
		assert(t2->cargo!=NULL);
		opSCAN(t2, relation)=opSCAN(t1, relation);
		break;
	default:
		assert(0);
	}
	t2->left=treedup(t1->left);
	t2->right=treedup(t1->right);
	return t2;
}

/*@}*/

/*@}*/
