/*
   ()   .
       .
@memo   
@name FileManager_Implementation
@author .. 
@version 1.0
@see DatabaseConfig Database
*/


#include "dbconfig.h"	/*     */
#include "dbgen.h"		/*    */
#include "filemgr.h"	/*  */
#include "errcode.h"	/*   */
#include "common.h"		/*     */
#include "debug.h"		/*    */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/*@{*/

/*
     
@name FileManager_Implementation_Types
@memo  
*/

/*@{*/

/*
@memo   
*/
typedef struct {
	int relation;
	char filename[12];
	FILE * f;
	tuple_t current;
	int processed;
} relation_t;


/*
@memo  
*/
typedef relation_t relationtab_t[RELATIONS_IN_DATABASE];

/*@}*/

/*
   
@name FileManager_Implementation_Variables
@memo  
*/

/*@{*/

/*
@memo  
*/
static relationtab_t RELATIONTAB;

/*
@memo  
*/
static dictionary_t * DICTIONARY;

/*
@memo   
*/
static int initflag=FALSE;

/*@}*/

/*
   
@name FileManager_Implementation_Functions
@memo  
*/

/*@{*/

/*
           .
          .
      .
@memo  
@param node  
@see fm_OpenRelation fm_CloseRelation 
*/
int fm_InitFileManager(int node)
{
	int i, j;

	DBG_INIT(node);
	START("fm_InitFileManager");
	if (initflag)
		return SUCCESS;
	if (node<0 || node>=FRAGMENTS_IN_RELATION)
		return EINVAL;
	if ((DICTIONARY=db_DatabaseDictionary())==NULL)
		return EINVAL;
	if (node!=DICTIONARY->whoami)
		return EINVAL;
	initflag=TRUE;
	for (i=0; i<RELATIONS_IN_DATABASE; i++) {
		RELATIONTAB[i].relation=NIL;
		strcpy(RELATIONTAB[i].filename, "");
		RELATIONTAB[i].f=NULL;
		RELATIONTAB[i].current=(tuple_t) calloc(ATTRIBUTES_IN_RELATION, sizeof(domain_t));
		assert(RELATIONTAB[i].current!=NULL);
		for (j=0; j<ATTRIBUTES_IN_RELATION; j++)
			RELATIONTAB[i].current[j]=NULL_VALUE;
		RELATIONTAB[i].processed=NIL;
   }
	STOP("fm_InitFileManager");
	return SUCCESS;
}

/*
      .
        .
@memo   
@param relation  
@return SUCCESS       
@see fm_CloseRelation fm_ResetRelation
*/
int fm_OpenRelation(int relation)
{
	int i;

	START("fm_OpenRelation");
	if (!initflag)
		return ENOINIT;
	if (relation<0 || relation>=RELATIONS_IN_DATABASE)
		return ESRCH;
	for (i=0; i<RELATIONS_IN_DATABASE; i++)
		if (RELATIONTAB[i].relation==relation) {
			RETURN_EINVAL("fm_OpenRelation");
			STOP("fm_OpenRelation");
			return EINVAL;
		}
	for (i=0; i<RELATIONS_IN_DATABASE; i++)
		if (RELATIONTAB[i].relation==NIL) {
			RELATIONTAB[i].relation=relation;
			strcpy(RELATIONTAB[i].filename, DICTIONARY->relations[relation].filename);
			RELATIONTAB[i].f=fopen(RELATIONTAB[i].filename, "rt");
			assert(RELATIONTAB[i].f!=NULL);
			RELATIONTAB[i].processed=0;
			RETURN_SUCCESS("fm_OpenRelation");
			STOP("fm_OpenRelation");
			return SUCCESS;
		}
	RETURN_EOVRFLOW("fm_OpenRelation");
	STOP("fm_OpenRelation");
	return EOVRFLOW;
}

/*
      .
        .
@memo   
@param relation  
@return SUCCESS       
@see ErrorCodes fm_OpenRelation fm_ResetRelation
*/
int fm_CloseRelation(int relation)
{
	int i, j;

	START("fm_CloseRelation");
	if (!initflag) {
		RETURN_ENOINIT("fm_CloseRelation");
		STOP("fm_CloseRelation");		
		return ENOINIT;
	}
	if (relation<0 || relation>=RELATIONS_IN_DATABASE) {
		RETURN_ESRCH("fm_CloseRelation");
		STOP("fm_CloseRelation");		
		return ESRCH;
	}
	for (i=0; i<RELATIONS_IN_DATABASE; i++)
		if (RELATIONTAB[i].relation==relation) {
			RELATIONTAB[i].relation=NIL;
			strcpy(RELATIONTAB[i].filename, "");
			fclose(RELATIONTAB[i].f);
	      for (j=0; j<ATTRIBUTES_IN_RELATION; j++)
				RELATIONTAB[i].current[j]=NULL_VALUE;
			RELATIONTAB[i].processed=NIL;
			RETURN_SUCCESS("fm_CloseRelation");
			STOP("fm_CloseRelation");		
			return SUCCESS;
		}
	RETURN_EINVAL("fm_CloseRelation");
	STOP("fm_CloseRelation");		
	return EINVAL;
}

/*
      , ..    
  .
        fm_NextTuple  fm_EOF   , :
fm_OpenRelation(myrelation);
fm_ResetRelation(myrelation);
while (!fm_EOF(mytupleptr=fm_NextTuple(myrelation))) {
	 ,   mytupleptr
}
fm_CloseRelation(myrelation);
             fm_OpenRelation.
@memo    
@param relation  
@return SUCCESS       
@see ErrorCodes db_InitDatabase fm_NextTuple fm_OpenRelation
*/
int fm_ResetRelation(int relation)
{
	int i, j;

	START("fm_ResetRelation");
	if (!initflag) {
		RETURN_ENOINIT("fm_ResetRelation");
		STOP("fm_ResetRelation");		
		return ENOINIT;
	}
	if (relation<0 || relation>=RELATIONS_IN_DATABASE)
		return ESRCH;
	for (i=0; i<RELATIONS_IN_DATABASE; i++)
		if (RELATIONTAB[i].relation==relation) {
			fclose(RELATIONTAB[i].f);
			RELATIONTAB[i].f=fopen(RELATIONTAB[i].filename, "rt");
			assert(RELATIONTAB[i].f!=NULL);
			for (j=0; j<ATTRIBUTES_IN_RELATION; j++)
				RELATIONTAB[i].current[j]=NULL_VALUE;
			RELATIONTAB[i].processed=0;
			RETURN_SUCCESS("fm_ResetRelation");
			STOP("fm_ResetRelation");		
			return SUCCESS;
		}
	RETURN_EINVAL("fm_ResetRelation");
	STOP("fm_ResetRelation");		
	return EINVAL;
}

/*
      .
        fm_ResetRelation  fm_EOF   , :
fm_OpenRelation(myrelation);
fm_ResetRelation(myrelation);
while (!EOF(mytupleptr=fm_NextTuple(myrelation))) {
	 ,   mytupleptr
}
fm_CloseRelation(myrelation);
@memo   
@param relation  
@return        NULL    
(,    )
@see fm_ResetRelation
*/
tuple_t fm_NextTuple(int relation)
{
	int i, j;

	START("fm_NextTuple");
	if (!initflag) {
		RETURN_NULL("fm_NextTuple");
		STOP("fm_NextTuple");		
		return NULL;
	}
	if (relation<0 || relation>=RELATIONS_IN_DATABASE) {
		RETURN_NULL("fm_NextTuple");
		STOP("fm_NextTuple");		
		return NULL;
	}
	for (i=0; i<RELATIONS_IN_DATABASE; i++)
		if (RELATIONTAB[i].relation==relation) {
			if (RELATIONTAB[i].processed==DICTIONARY->relations[relation].tuples) {
				for (j=0; j<ATTRIBUTES_IN_RELATION; j++)
					RELATIONTAB[i].current[j]=EOFmarker;
				RETURN_EOF("fm_NextTuple");
				STOP("fm_NextTuple");
				return RELATIONTAB[i].current;
			}
			for (j=0; j<ATTRIBUTES_IN_RELATION; j++)
				fscanf(RELATIONTAB[i].f, "%d\t", &RELATIONTAB[i].current[j]);
			fscanf(RELATIONTAB[i].f, "\n");
	        RELATIONTAB[i].processed++;
			RETURN_notEOF("fm_NextTuple");
			STOP("fm_NextTuple");
			return RELATIONTAB[i].current;
		}
	RETURN_NULL("fm_NextTuple");
	STOP("fm_NextTuple");		
	return NULL;
}

/*@}*/

/*@}*/

