/*
      
         .
@memo   
@name  PlanGenerator
@author .. , .. 
@version 1.0
@see Compiler
*/

#include "lineplan.h"	/*  */
#include "dbgen.h"		/*    */
#include "errcode.h"	/*   */
#include "common.h"		/*     */
#include "bracket.h"	/*   */
#include "filemgr.h"	/*   */
#include "debug.h"		/*    */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>


static void Generate(treenode_t *); /*    */

/*
@memo     
@param node    
@param tuple   
*/
#define RESCPY(node, tuple)	memcpy(node->result, tuple, node->ary*sizeof(domain_t))

/*@{*/

/*
    
@name Compiler_Implementation_Types
@memo   
@see Compiler_Functions Compiler_Constants
*/

/*@{*/

/*
@memo  
*/
static dictionary_t * DICTIONARY;

/*
@memo   
*/
static int initflag=FALSE;

/*@}*/

/**
     
@name PlanGenerator_Types
@memo    
@see PlanGenerator_Functions PlanGenerator_Constants
*/

/*@{*/



/*@}*/


/**
     
@name PlanGenerator_Functions
@memo    
@see PlanGenerator_Types PlanGenerator_Constants
*/

/*@{*/


/**
        
         .
       .
@memo    
@param node	   	
@return SUCCESS       
@see lp_GeneratePlan
*/
int lp_InitPlanGenerator(int node)
{
	DBG_INIT(node);
	START("lp_InitPlanGenerator");
	if (initflag)
		return SUCCESS;
	if (node<0 || node>=FRAGMENTS_IN_RELATION)
		return EINVAL;
	if ((DICTIONARY=db_DatabaseDictionary())==NULL)
		return EINVAL;
	if (node!=DICTIONARY->whoami)
		return EINVAL;
	initflag=TRUE;   
	STOP("lp_InitPlanGenerator");
	return SUCCESS;
}

/**
          
      .
@memo  
@param     
@return SUCCESS       
@see cp_CompileQuery
*/
int lp_GeneratePlan(treenode_t * root)
{
	START("lp_GeneratePlan");
	if (!initflag)
		return ENOINIT;
	Generate(root);
	STOP("lp_GeneratePlan");
	return SUCCESS;
}

/*
     RESTRICT.     
  " ".
@memo   RESTRICT
@param node     
@return        , 
  ;     --   -EOF.
@see EOFtuple execute_fn_t Bracket ResetRESTRICT ExecuteJOIN_NestedLoops ExecuteSCAN ExecuteSPLIT ExecuteMERGE ExecuteGATHER ExecuteSCATTER
*/
tuple_t ExecuteRESTRICT(treenode_t * node)
{
	tuple_t t;

	START("ExecuteRESTRICT");
	while (!node->left->isEOF) {
		t=br_Next(node->left);
		switch (opRESTRICT(node, theta)) {
		case EQUAL: 
			if (t[opRESTRICT(node, attribute)]==opRESTRICT(node, literal)) {
				RESCPY(node, t);
				return node->result;
			}
			break;
		case NOT_EQUAL:
			if (t[opRESTRICT(node, attribute)]!=opRESTRICT(node, literal)) {
				RESCPY(node, t);
				return node->result;
			}
			break;
		case GREATER:
			if (t[opRESTRICT(node, attribute)]>opRESTRICT(node, literal)) {
				RESCPY(node, t);
				return node->result;
			}
			break;
		case LESS:
			if (t[opRESTRICT(node, attribute)]<opRESTRICT(node, literal)) {
				node->card++;
				RESCPY(node, t);
				return node->result;
			}
			break;
		case GREATER_OR_EQUAL:
			if (t[opRESTRICT(node, attribute)]>=opRESTRICT(node, literal)) {
				RESCPY(node, t);
				return node->result;
			}
			break;
		case LESS_OR_EQUAL:
			if (t[opRESTRICT(node, attribute)]<=opRESTRICT(node, literal)) {
				RESCPY(node, t);
				return node->result;
			}
			break;
		}
	}
	RESCPY(node, EOFtuple);
	STOP("ExecuteRESTRICT");
	return node->result;
}

/*
   RESTRICT     ,     
  : return SUCCESS;
@memo   RESTRICT
@return SUCCESS       
@see reset_fn_t ExecuteRESTRICT
*/
int ResetRESTRICT(treenode_t * node)
{
	START("ResetRESTRICT");
	node->result=(tuple_t) calloc(node->ary, sizeof(domain_t));
	assert(node->result!=NULL);
	STOP("ResetRESTRICT");
	return SUCCESS;
}

/*
     JOIN   .   
    " ".

.
  JOIN -   (,   ,   -, 
  .).
@memo   JOIN   
@param node     
@return        , 
  ;     --   -EOF.
@see execute_fn_t Bracket ExecuteRESTRICT ExecuteSCAN ExecuteSPLIT ExecuteMERGE ExecuteGATHER ExecuteSCATTER
*/
tuple_t ExecuteJOIN_NestedLoops(treenode_t * node)
{
#define L	node->left->result
#define R	node->right->result

	int i;

	START("ExecuteJOIN_NestedLoops");

	if (node->left->isBOF) {
		WATCH(("ExecuteJOIN_NestedLoops: left son is BOF, doing Next of left son"));
		br_Next(node->left);
	}
	if (node->left->isEOF) {
		RESCPY(node, EOFtuple);
		RETURN_EOF("ExecuteJOIN_NestedLoops");
		STOP("ExecuteJOIN_NestedLoops");
		return node->result;
	}

	do {
		WATCH(("ExecuteJOIN_NestedLoops: doing Next of right son"));
		br_Next(node->right);
		if (node->right->isEOF) {
			WATCH(("ExecuteJOIN_NestedLoops: EOF of right son reached, do Reset, then Next of right son"));
			br_Reset(node->right);
			br_Next(node->right);
			if (node->right->isEOF) {
				WATCH(("ExecuteJOIN_NestedLoops: EOF of right son reached, must return EOF"));
				RESCPY(node, EOFtuple);
				RETURN_EOF("ExecuteJOIN_NestedLoops");
				STOP("ExecuteJOIN_NestedLoops");
				return node->result;	
			}
			WATCH(("ExecuteJOIN_NestedLoops: doing Next of left son"));
			br_Next(node->left);
			if (node->left->isEOF) {
				WATCH(("ExecuteJOIN_NestedLoops: EOF of left son reached, must return EOF"));
				RESCPY(node, EOFtuple);
				RETURN_EOF("ExecuteJOIN_NestedLoops");
				STOP("ExecuteJOIN_NestedLoops");
				return node->result;
			}
		}
	} while (L[opJOIN(node, attribute)]!=R[opJOIN(node, attribute)]);

	WATCH(("ExecuteJOIN_NestedLoops: make result tuple"));
	for (i=0; i<node->left->ary; i++)
		node->result[i]=L[i];
	for (i=0; i<opJOIN(node, attribute); i++)
		node->result[node->left->ary+i]=R[i];
	for (i=opJOIN(node, attribute)+1; i<node->right->ary; i++)
		node->result[node->left->ary+i-1]=R[i];

	for (i=0; i<node->ary; i++)
		WATCH(("ExecuteJOIN_NestedLoops\tnode->result{%d}==%d", i, node->result[i]));

	RETURN_notEOF("ExecuteJOIN_NestedLoops");
	STOP("ExecuteJOIN_NestedLoops");
	return node->result;

#undef L
#undef R
}

/*
   JOIN     ,     
  : return SUCCESS;
@memo   JOIN
@return SUCCESS       
@see reset_fn_t ExecuteJOIN_NestedLoops
*/
int ResetJOIN(treenode_t * node)
{
	START("ResetJOIN");
	node->result=(tuple_t) calloc(node->ary, sizeof(domain_t));
	assert(node->result!=NULL);
	STOP("ResetJOIN");
	return SUCCESS;
}

/*
     SCAN       
 .
@memo   SCAN
@param node     
@return        , 
  ;     --   -EOF.
@see execute_fn_t FileManager ExecuteRESTRICT ExecuteJOIN_NestedLoops ExecuteSPLIT ExecuteMERGE ExecuteGATHER ExecuteSCATTER
*/
tuple_t ExecuteSCAN(treenode_t * node)
{
	START("ExecuteSCAN");
	node->result=fm_NextTuple(opSCAN(node, relation));
	assert(node->result!=NULL);
	STOP("ExecuteSCAN");
	return node->result;
}

/*
      SCAN      
    .
@memo   SCAN
@return SUCCESS       
@see reset_fn_t ExecuteSCAN
*/
int ResetSCAN(treenode_t * node)
{
	int res;

	START("ResetSCAN");
	node->result=(tuple_t) calloc(node->ary, sizeof(domain_t));
	assert(node->result!=NULL);
	res=fm_ResetRelation(opSCAN(node, relation));
	assert(res==SUCCESS);
	STOP("ResetSCAN");
	return SUCCESS;
}

/*
@memo  
@param root     
@see lp_GeneratePlan
*/
static void Generate(treenode_t * root)
{
	START("Generate");
	if (root==NULL) {
		STOP("Generate");
		return;
	}
	switch (root->operation) {
	case JOIN: 
		root->executefn=(execute_fn_t) ExecuteJOIN_NestedLoops;
		root->resetfn=(reset_fn_t) ResetJOIN;
		break;
	case RESTRICT:
		root->executefn=(execute_fn_t) ExecuteRESTRICT;
		root->resetfn=(reset_fn_t) ResetRESTRICT;
		break;
	case SCAN:
		root->executefn=(execute_fn_t) ExecuteSCAN;
		root->resetfn=(reset_fn_t) ResetSCAN;
		break;
	default:
		assert(0);
	}
	Generate(root->left);
	Generate(root->right);
}

/*@}*/

/*@}*/
