/*
      
          .
     .

.
1.           EXCHANGE. 
        5 ( 5.3, . 163  ) 
    ..           
    : . ... . .-. : 05.13.18 /   
   . - , 2003. - 247 . [http://sok.susu.ru/dissertation/dissert.pdf], 16.06.2003.
2.       :      JOIN    
     SCAN.

.
    (    ).
@memo  
@name  Parallelizer_Implementation
@author (c) 2004 .. , .. 
@version 1.0
@see Compiler Database
*/

#include "parallel.h"	/*  */
#include "errcode.h"	/*   */
#include "common.h"		/*     */
#include "dbgen.h"		/*    */
#include "bracket.h"	/*   */
#include "msgmgr.h"		/*   */
#include "debug.h"		/*    */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


static void InsertEXCHANGE(treenode_t *); /*   EXCHANGE */
static void ExpandEXCHANGE(treenode_t *); /*   EXCHANGE (=MERGE+GATHER+SPLIT+SCATTER) */


/*
@memo     
@param node    
@param tuple   
*/
#define RESCPY(node, tuple)	memcpy(node->result, tuple, node->ary*sizeof(domain_t))

/*
        -    . 
 ,              -.
@memo      
@param whoami   
@param portEXCHANGE    EXCHANGE,       
@param dest  - 
@return     
@see
*/
#define PORT_SEND(whoami, portEXCHANGE, dest)	(whoami)+NODES_IN_CLUSTER*(dest)+(portEXCHANGE)*10

/*
        -    . 
 ,              -.
@memo      
@param whoami   
@param portEXCHANGE    EXCHANGE,       
@param dest  - 
@return     
@see
*/
#define PORT_RECV(whoami, portEXCHANGE, source)	(whoami)*NODES_IN_CLUSTER+(source)+(portEXCHANGE)*10


/*@{*/

/*
     
@name Parallelizer_Implementation_Types
@memo    
@see Parallelizer_Implementation_Functions Parallelizer_Implementation_Constants
*/

/*@{*/



/*@}*/

/*
     
@name Parallelizer_Implementation_Types
@memo   
@see Parallelizer_Implementation_Functions Parallelizer_Implementation_Constants
*/

/*@{*/

/*
@memo  
*/
static dictionary_t * DICTIONARY;

/*
@memo   
*/
static int initflag=FALSE;

/*
@memo  ""
*/
static int port;

/*@}*/

/*
    
@name Parallelizer_Implementation_Functions
@memo   
@see Parallelizer_Implementation_Types Parallelizer_Implementation_Constants
*/

/*@{*/


/*
       
         .
       .
@memo   
@param node	   	
@return SUCCESS       
@see pl_ParallelizeQuery
*/
int pl_InitParallelizer(int node)
{
	DBG_INIT(node);
	START("pl_InitParallelizer");
	if (initflag)
		return SUCCESS;
	if (node<0 || node>=FRAGMENTS_IN_RELATION)
		return EINVAL;
	if ((DICTIONARY=db_DatabaseDictionary())==NULL)
		return EINVAL;
	if (node!=DICTIONARY->whoami)
		return EINVAL;
	initflag=TRUE;
	STOP("pl_InitParallelizer");
	return SUCCESS;
}

/*
        
    .     EXCHANGE 
" "    .   0   
 EXCHANGE,     .
@memo  
@param root        
@return SUCCESS       
@see treenode_t AgentManager am_MakeAgent
*/
int pl_ParallelizeQuery(treenode_t * root)
{
	START("pl_ParallelizeQuery");
	if (!initflag)
		return ENOINIT;
	port=1; /*   !! */
	InsertEXCHANGE(root);
	ExpandEXCHANGE(root);
	STOP("pl_ParallelizeQuery");	
	return SUCCESS;
}

/*
        
    .
@memo  
@param root        
*/
static void InsertEXCHANGE(treenode_t * root)
{
	treenode_t *lnode=NULL, *rnode=NULL;

	START("InsertEXCHANGE");
	if (root==NULL) {
		STOP("InsertEXCHANGE");	
		return;
	}
	switch (root->operation) {
	case SCAN: 
		root->fragmentattr=DICTIONARY->relations[opSCAN(root, relation)].fragmentattr;
		root->fragmentfn=DICTIONARY->relations[opSCAN(root, relation)].fragmentfn;
		root->left=root->right=NULL;
		break;
	case RESTRICT:
		if (root->left->fragmentattr==NIL)
			InsertEXCHANGE(root->left);
		root->fragmentattr=root->left->fragmentattr;
		root->fragmentfn=root->left->fragmentfn;
		break;
	case JOIN:
		root->fragmentattr=opJOIN(root, attribute);
		if (root->left->fragmentattr==NIL)
			InsertEXCHANGE(root->left);
		if (root->fragmentattr!=root->left->fragmentattr) {
			lnode=(treenode_t *) calloc(1, sizeof(treenode_t));
			assert(lnode!=NULL);
			lnode->operation=EXCHANGE;
			lnode->ary=root->left->ary;
			lnode->cargo=(opEXCHANGE_t *) calloc(1, sizeof(opEXCHANGE_t));
			assert(lnode->cargo!=NULL);
			opEXCHANGE(lnode, port)=port++;
			lnode->right=NULL;
			lnode->left=root->left;
			lnode->fragmentattr=root->fragmentattr;
			lnode->fragmentfn=root->fragmentfn;
		}
		if (root->right->fragmentattr==NIL)
			InsertEXCHANGE(root->right);
		if (root->fragmentattr!=root->right->fragmentattr) {
			rnode=(treenode_t *) calloc(1, sizeof(treenode_t));
			assert(rnode!=NULL);
			rnode->operation=EXCHANGE;
			rnode->ary=root->right->ary;
			rnode->cargo=(opEXCHANGE_t *) calloc(1, sizeof(opEXCHANGE_t));
			assert(rnode->cargo!=NULL);
			opEXCHANGE(rnode, port)=port++;
			rnode->right=NULL;
			rnode->left=root->right;
			rnode->fragmentattr=root->fragmentattr;
			rnode->fragmentfn=root->fragmentfn;
		}
		if (root->left->fragmentfn!=root->right->fragmentfn)
			root->fragmentfn=db_FragmentByDefault;
		else
			root->fragmentfn=root->left->fragmentfn;
		if (lnode!=NULL) {
			opEXCHANGE(lnode, distributefn)=root->fragmentfn;
			root->left=lnode;
		}
		if (rnode!=NULL) {
			opEXCHANGE(rnode, distributefn)=root->fragmentfn;
			root->right=rnode;
		}
		break;
	default:
		assert(0);
	}
}

/*
        
    .
@memo   EXCHANGE  (MERGE+GATHER+SPLIT+SCATTER)
@param root        
*/
static void ExpandEXCHANGE(treenode_t * root)
{
	treenode_t *Gather, *Split, *Scatter;
	int port;
	distribute_fn_t distributefn;

	START("ExpandEXCHANGE");
	if (root==NULL) {
		STOP("ExpandEXCHANGE");	
		return;
	}
	if (root->operation==EXCHANGE) {
		port=opEXCHANGE(root, port);
		distributefn=opEXCHANGE(root, distributefn);

		root->operation=MERGE;
		free(root->cargo);
		root->cargo=(opMERGE_t *) calloc(1, sizeof(opMERGE_t));
		assert(root->cargo!=NULL);
		root->executefn=(execute_fn_t) ExecuteMERGE;
		root->resetfn=(reset_fn_t) ResetMERGE;

		Gather=(treenode_t *) calloc(1, sizeof(treenode_t));
		assert(Gather!=NULL);
		Gather->operation=GATHER;
		Gather->cargo=(opGATHER_t *) calloc(1, sizeof(opGATHER_t));
		assert(Gather->cargo!=NULL);
		opGATHER(Gather, port)=port;
		Gather->executefn=(execute_fn_t) ExecuteGATHER;
		Gather->resetfn=(reset_fn_t) ResetGATHER;
		Gather->left=Gather->right=NULL;
		Gather->ary=root->ary;
		Gather->card=root->card;
		Gather->fragmentattr=root->fragmentattr;
		Gather->fragmentfn=root->fragmentfn;
		
		Split=(treenode_t *) calloc(1, sizeof(treenode_t));
		assert(Split!=NULL);
		Split->operation=SPLIT;
		Split->cargo=(opSPLIT_t *) calloc(1, sizeof(opSPLIT_t));
		assert(Split->cargo!=NULL);
		opSPLIT(Split, distributefn)=distributefn;
		Split->executefn=(execute_fn_t) ExecuteSPLIT;
		Split->resetfn=(reset_fn_t) ResetSPLIT;
		Split->ary=root->ary;
		Split->card=root->card;
		Split->fragmentattr=root->fragmentattr;
		Split->fragmentfn=root->fragmentfn;
	
		Scatter=(treenode_t *) calloc(1, sizeof(treenode_t));
		assert(Scatter!=NULL);
		Scatter->operation=SCATTER;
		Scatter->cargo=(opSCATTER_t *) calloc(1, sizeof(opSCATTER_t));
		assert(Scatter->cargo!=NULL);
		opSCATTER(Scatter, port)=port;
		Scatter->executefn=(execute_fn_t) ExecuteSCATTER;
		Scatter->resetfn=(reset_fn_t) ResetSCATTER;
		Scatter->left=Scatter->right=NULL;
		Scatter->ary=root->ary;
		Scatter->card=root->card;
		Scatter->fragmentattr=root->fragmentattr;
		Scatter->fragmentfn=root->fragmentfn;

		root->right=Split;
		Split->left=root->left;
		Split->right=Scatter;
		root->left=Gather;
	}
	ExpandEXCHANGE(root->left);
	ExpandEXCHANGE(root->right);
}

/*
     MERGE (  EXCHANGE). 
      " "   
 .
@memo   MERGE
@return       MERGE  , 
   --     NULL   
*/
tuple_t ExecuteMERGE(treenode_t * node)
{
	tuple_t tl=NULL, tr=NULL;

	START("ExecuteMERGE");
	WATCH(("ExecuteMERGE (%d)", opGATHER(node->left, port)));
	while (TRUE) {
		++opMERGE(node, cnt);
		opMERGE(node, cnt)%=2;
		switch (opMERGE(node, cnt)) {
		case 0:
			WATCH(("ExecuteMERGE (%d)\todd case, try to get tuple from SPLIT son first", opGATHER(node->left, port)));
			tr=br_Next(node->right);
			if (!IsEOF(tr) && tr!=NULL) {
				WATCH(("ExecuteMERGE (%d)\treturn non-EOF tuple from SPLIT son", opGATHER(node->left, port)));
				RESCPY(node, tr);
				WATCH(("ExecuteMERGE (%d)", opGATHER(node->left, port)));
				STOP("ExecuteMERGE");
				return node->result;
			} 
			tl=br_Next(node->left);
			if (!IsEOF(tl) && tl!=NULL) {
				WATCH(("ExecuteMERGE\treturn non-EOF tuple from GATHER son"));
				RESCPY(node, tl);
				WATCH(("ExecuteMERGE (%d)", opGATHER(node->left, port)));
				STOP("ExecuteMERGE");
				return node->result;
			}
			break;
		case 1:
			WATCH(("ExecuteMERGE (%d)\teven case, try to get tuple from GATHER son first", opGATHER(node->left, port)));
			tl=br_Next(node->left);
			if (!IsEOF(tl) && tl!=NULL) {
				WATCH(("ExecuteMERGE (%d)\treturn non-EOF tuple from GATHER son", opGATHER(node->left, port)));
				RESCPY(node, tl);
				WATCH(("ExecuteMERGE (%d)", opGATHER(node->left, port)));
				STOP("ExecuteMERGE");
				return node->result;
			} 
			tr=br_Next(node->right);
			if (!IsEOF(tr) && tr!=NULL) {
				WATCH(("ExecuteMERGE (%d)\treturn non-EOF tuple from SPLIT son", opGATHER(node->left, port)));
				RESCPY(node, tr);
				WATCH(("ExecuteMERGE (%d)", opGATHER(node->left, port)));
				STOP("ExecuteMERGE");
				return node->result;
			}
			break;
		default:
			assert(0);
		}
		if (IsEOF(tl) && IsEOF(tr))
			break;
	} 
	RESCPY(node, EOFtuple);
	RETURN_EOF("ExecuteMERGE");
	WATCH(("ExecuteMERGE (%d)", opGATHER(node->left, port)));
	STOP("ExecuteMERGE");
	return node->result;
}

/*
   MERGE      .
@memo   MERGE
@param node     
@return SUCCESS       
@see reset_fn_t ExecuteMERGE
*/
int ResetMERGE(treenode_t * node)
{
	START("ResetMERGE");
	WATCH(("ResetMERGE (%d)", opGATHER(node->left, port)));
	node->result=(tuple_t) calloc(node->ary, sizeof(domain_t));
	assert(node->result!=NULL);
	opMERGE(node, cnt)=0;
	WATCH(("ResetMERGE (%d)", opGATHER(node->left, port)));
	STOP("ResetMERGE");
	return SUCCESS;
}

/*
     SPLIT (  EXCHANGE). 
      " "   
 .
@memo   SPLIT
@return       SPLIT  , 
   --     NULL   
*/
tuple_t ExecuteSPLIT(treenode_t * node)
{
	int i, dest; 

	START("ExecuteSPLIT");
	WATCH(("ExecuteSPLIT (%d)", opSCATTER(node->right, port)));

	if (opSPLIT(node, cntEOF)==NODES_IN_CLUSTER-1) {
		WATCH(("ExecuteSPLIT\tstill return EOF"));
		RESCPY(node, EOFtuple);
		RETURN_EOF("ExecuteSPLIT");
		WATCH(("ExecuteSPLIT (%d)", opSCATTER(node->right, port)));
		STOP("ExecuteSPLIT");
		return node->result;
	}

	if (!opSPLIT(node, breakover)) { /*      SPLIT */
		WATCH(("ExecuteSPLIT (%d)\thad not breakover, doing Next for left son", opSCATTER(node->right, port)));
		node->result=br_Next(node->left);
		if (IsEOF(node->result)) {
			WATCH(("ExecuteSPLIT (%d)\t got EOF, now waiting for EOFs from other nodes"));
			while (opSPLIT(node, cntEOF)!=NODES_IN_CLUSTER-1) 
				for (i=0; i<NODES_IN_CLUSTER; i++) {
					if (i==DICTIONARY->whoami)
						continue;
					if (!opSCATTER(node->right, ready[i]))
						if (mm_IsDone(opSCATTER(node->right, midSend[i]))!=FALSE) {
							opSCATTER(node->right, ready[i])=TRUE;
							memcpy(opSCATTER(node->right, bufSend[i]), node->result, node->ary*sizeof(domain_t));
							opSPLIT(node, cntEOF)++;
						}
				}
			WATCH(("ExecuteSPLIT (%d): doing Next for right son (SCATTER)", opSCATTER(node->right, port)));
			br_Next(node->right);
			RETURN_EOF("ExecuteSPLIT");
			WATCH(("ExecuteSPLIT (%d)", opSCATTER(node->right, port)));
			STOP("ExecuteSPLIT");
			return EOFtuple;
		}
	} else
		WATCH(("ExecuteSPLIT: had breakover"));

	WATCH(("ExecuteSPLIT: node->fragmentattr==%d", node->fragmentattr));
	dest=opSPLIT(node, distributefn)(node->result, node->fragmentattr);
	WATCH(("ExecuteSPLIT (%d)\t got tuple for %d node", opSCATTER(node->right, port), dest));
	assert(dest>=0 && dest<NODES_IN_CLUSTER);
	if (dest==DICTIONARY->whoami) {
		WATCH(("ExecuteSPLIT (%d)\t got my tuple", opSCATTER(node->right, port)));
		RETURN_notEOF("ExecuteSPLIT");
		WATCH(("ExecuteSPLIT (%d)", opSCATTER(node->right, port)));
		STOP("ExecuteSPLIT");
		return node->result;
	}
	if (mm_IsDone(opSCATTER(node->right, midSend[dest]))!=FALSE) {
		WATCH(("ExecuteSPLIT (%d)\tSCATTER can send tuple for %d node", opSCATTER(node->right, port), dest));
		opSPLIT(node, breakover)=FALSE;
		opSCATTER(node->right, ready[dest])=TRUE;
		memcpy(opSCATTER(node->right, bufSend[dest]), node->result, node->ary*sizeof(domain_t));
	} else {
		WATCH(("ExecuteSPLIT (%d)\t SCATTER can't send tuple for %d node, it's breakover", opSCATTER(node->right, port), dest));
		opSPLIT(node, breakover)=TRUE;	
	}

	WATCH(("ExecuteSPLIT (%d): doing Next for right son (SCATTER)", opSCATTER(node->right, port)));
	br_Next(node->right);
	RETURN_NULL("ExecuteSPLIT");
	WATCH(("ExecuteSPLIT (%d)", opSCATTER(node->right, port)));
	STOP("ExecuteSPLIT");
	return NULL;	
	
}

/*
   SPLIT      .
@memo   SPLIT
@param node     
@return SUCCESS       
@see reset_fn_t ExecuteSPLIT
*/
int ResetSPLIT(treenode_t * node)
{
	START("ResetSPLIT");
	WATCH(("ResetSPLIT (%d)", opSCATTER(node->right, port)));

	node->result=(tuple_t) calloc(node->ary, sizeof(domain_t));
	assert(node->result!=NULL);
	opSPLIT(node, breakover)=FALSE;
	opSPLIT(node, cntEOF)=0;
	
	WATCH(("ResetSPLIT (%d)", opSCATTER(node->right, port)));	
	STOP("ResetSPLIT");
	return SUCCESS;
}

/*
     GATHER (  EXCHANGE). 
      " "   
 .
@memo   GATHER
@return       GATHER  , 
   --     NULL   
*/
tuple_t ExecuteGATHER(treenode_t * node)
{
	int i, allempty, portRecv;
#ifndef NDEBUG
	int cnt;
#endif

	START("ExecuteGATHER");
	WATCH(("ExecuteGATHER (%d)", opGATHER(node, port)));

	if (opGATHER(node, cntEOF)==NODES_IN_CLUSTER-1) {
		WATCH(("ExecuteGATHER\tall buffers still EOF"));
		RESCPY(node, EOFtuple);
		RETURN_EOF("ExecuteGATHER");
		STOP("ExecuteGATHER");
		return node->result;
	}

	allempty=TRUE;
	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==DICTIONARY->whoami)
			continue;
		WATCH(("ExecuteGATHER (%d)\tCheck recv operation from %d node (id==%d)", opGATHER(node, port), i, opGATHER(node,midRecv[i])));

		if (mm_IsDone(opGATHER(node, midRecv[i]))==TRUE) {
			opGATHER(node, midRecv[i])=NIL;
#ifndef NDEBUG
			for (cnt=0; cnt<node->ary; cnt++)
				WATCH(("ExecuteGATHER (%d): bufRecv[%d]: {%d}==%d", opGATHER(node, port), i, cnt, opGATHER(node, bufRecv[i])[cnt]));
#endif
			allempty=FALSE;
			if (IsEOF(opGATHER(node, bufRecv[i]))) {
				WATCH(("ExecuteGATHER (%d)\tGATHER received EOF tuple from node %d", opGATHER(node, port), i));
				opGATHER(node, cntEOF)++;
				if (opGATHER(node, cntEOF)==NODES_IN_CLUSTER-1) {
					WATCH(("ExecuteGATHER (%d)\tall buffers are EOF", opGATHER(node, port)));
					RESCPY(node, EOFtuple);
					RETURN_EOF("ExecuteGATHER");
					STOP("ExecuteGATHER");
					return node->result;
				}
			} else { 
				WATCH(("ExecuteGATHER (%d)\tGATHER received non-EOF tuple from node %d", opGATHER(node, port), i));
				RESCPY(node, opGATHER(node, bufRecv[i]));
				/*     */
				portRecv=PORT_RECV(DICTIONARY->whoami, opGATHER(node, port), i);
				WATCH(("ExecuteGATHER (%d)\trecv tuple from %d node using port %d", opGATHER(node, port), i, portRecv));
				opGATHER(node, midRecv[i])=mm_RunRecv(portRecv, i, opGATHER(node, bufRecv[i]), node->ary*sizeof(domain_t));
				assert(opGATHER(node, midRecv[i])>=0);
				STOP("ExecuteGATHER");
				return node->result;
			}
		}
	}
	if (allempty) {
		WATCH(("ExecuteGATHER (%d)\tall buffers are empty", opGATHER(node, port)));
		RETURN_NULL("ExecuteGATHER");
		STOP("ExecuteGATHER");
		return NULL;
	}
	RETURN_NULL("ExecuteGATHER");	
	STOP("ExecuteGATHER");
	return NULL;
}

/*
  GATHER         
    ()     . 
        .
@memo   GATHER
@param node     
@return SUCCESS       
@see MessageManager reset_fn_t ExecuteGATHER 
*/
int ResetGATHER(treenode_t * node)
{
	int i, portRecv;

	START("ResetGATHER");
	WATCH(("ResetGATHER (%d)", opGATHER(node, port)));

	node->result=(tuple_t) calloc(node->ary, sizeof(domain_t));
	assert(node->result!=NULL);
	opGATHER(node, cntEOF)=0;
			
	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==DICTIONARY->whoami) {
			opGATHER(node, bufRecv[i])=NULL;
			opGATHER(node, midRecv[i])=NIL;
			continue;
		}
		opGATHER(node, bufRecv[i])=(tuple_t) calloc(node->ary, sizeof(domain_t));
		assert(opGATHER(node, bufRecv[i])!=NULL);
		portRecv=PORT_RECV(DICTIONARY->whoami, opGATHER(node, port), i);
		WATCH(("ResetGATHER (%d)\trecv tuple from %d node using port %d", opGATHER(node, port), i, portRecv));
		opGATHER(node, midRecv[i])=mm_RunRecv(portRecv, i, opGATHER(node, bufRecv[i]), node->ary*sizeof(domain_t));
		assert(opGATHER(node, midRecv[i])>=0);
	}

	WATCH(("ResetGATHER (%d)", opGATHER(node, port)));
	STOP("ResetGATHER");
	return SUCCESS;
}

/*
     SCATTER (  EXCHANGE). 
      " "   
 .
@memo   SCATTER
@return       SCATTER  , 
   --     NULL   
*/
tuple_t ExecuteSCATTER(treenode_t * node)
{
	int i, portSend;

	START("ExecuteSCATTER");
	WATCH(("ExecuteSCATTER (%d)", opSCATTER(node, port)));

	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==DICTIONARY->whoami)
			continue;
		if (opSCATTER(node, ready[i])) {
			portSend=PORT_SEND(DICTIONARY->whoami, opSCATTER(node, port), i);
			WATCH(("ExecuteSCATTER (%d)\tsend tuple to %d node using port %d", opSCATTER(node, port), i, portSend));
			opSCATTER(node, midSend[i])=mm_RunSend(portSend, i, opSCATTER(node, bufSend[i]), node->ary*sizeof(domain_t));
			assert(opSCATTER(node, midSend[i])>=0);
			opSCATTER(node, ready[i])=FALSE;
		}
	}
	RETURN_NULL("ExecuteSCATTER");
	WATCH(("ExecuteSCATTER (%d)", opSCATTER(node, port)));	
	STOP("ExecuteSCATTER");
	return NULL;
}

/*
   SCATTER      .
@memo   SCATTER
@param node     
@return SUCCESS       
@see reset_fn_t ExecuteSCATTER
*/
int ResetSCATTER(treenode_t * node)
{
	int i;

	START("ResetSCATTER");
	WATCH(("ResetSCATTER (%d)", opSCATTER(node, port)));

	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==DICTIONARY->whoami) {
			opSCATTER(node, bufSend[i])=NULL;
			opSCATTER(node, midSend[i])=NIL;
			opSCATTER(node, ready[i])=NIL;
			continue;
		}
		opSCATTER(node, bufSend[i])=(tuple_t) calloc(node->ary, sizeof(domain_t));
		assert(opSCATTER(node, bufSend[i])!=NULL);
		opSCATTER(node, midSend[i])=NIL;
		opSCATTER(node, ready[i])=FALSE;
	}
	WATCH(("ResetSCATTER (%d)", opSCATTER(node, port)));
	STOP("ResetSCATTER");
	return SUCCESS;
}

/*@}*/

/*@}*/
